#include <avr/io.h>
#include <avr/interrupt.h>   // for cli()
#include "delay_x.h"
#include "usb_serial.h"

class SinglePort {
public:
  typedef unsigned char valuetype;
  static inline void set(valuetype value) {
    PORTD = value;
//    _delay_us(10);
    _delay_us(6);
  }
  static inline void setinv(valuetype value) {
    set(~value);
  }
  static inline void high() { set(0xff); }
  static inline void low() { set(0); }
};

template<class Port>
void SendFrame(typename Port::valuetype *v) {
  unsigned char oldSREG = SREG;
  cli();
  Port::high();
  for (int i = 0; i < 26; i++) {
    Port::low();
    Port::setinv(v[i]);
    Port::high();
  }
  Port::low();
  // We can tolerate interrupts now.
  SREG = oldSREG;
  Port::low();
  Port::low();
}

#define CPU_PRESCALE(n) (CLKPR = 0x80, CLKPR = (n))

unsigned char buffer[26];

extern "C" int main(void) {
  CPU_PRESCALE(0);
  usb_init();
  DDRD = 0xff;

  while (1) {
    while (usb_serial_getchar() != 0xEE);
    for (int i = 0; i < 26; i++) {
      buffer[i] = usb_serial_getchar();
    }
    SendFrame<SinglePort>(buffer);
  }
}
