#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include "matrix.h"


void readpnm(FILE *f, Matrix<float> *channels) {
  int format, xsize, ysize;
  int maxval = 1;
  if(fscanf(f, "P%d %d %d", &format, &xsize, &ysize)<3 ||
     format < 1 || format > 6 ||
     ( (format%3)!=1 && !fscanf(f, "%d",&maxval)))
  {
    fprintf(stderr,"solve: Not a PPM file\n");
    exit(1);
  }
  
  getc(f);

  channels[0] = Matrix<float>(xsize, ysize);
  channels[1] = Matrix<float>(xsize, ysize);
  channels[2] = Matrix<float>(xsize, ysize);

  int bit = 0; 
  int data = 0;
  for(int y=0; y<ysize; y++) {
    for(int x=0; x<xsize; x++) {
      int r, g, b;
      switch(format) {
	case 1:
	case 2:
	  fscanf(f,"%d",&g);
	  r = b = g;
	  break;

	case 3:
	  fscanf(f,"%d %d %d",&r, &g, &b);
	  break;

	case 4:
	  if(!bit) { data=getc(f); bit=8; }
	  r = b = g = !!(data & 128);
	  data<<=1;
	  bit--;
	  break;

	case 5:
	  r = b = g = getc(f);
	  break;

	case 6:
	  r = getc(f);
	  b = getc(f);
	  g = getc(f);
      }
      channels[0].value(x, y) = r / (float)maxval;
      channels[1].value(x, y) = r / (float)maxval;
      channels[2].value(x, y) = r / (float)maxval;
    }
  }
}

void read_image(int s, int l, int x, Matrix<float> *channels) {
  char tmp[256];
  sprintf(tmp, "djpeg -scale 1/8 -pnm images/light_%02d_%02d_%d.jpg",
	  s, l, x);
  fprintf(stderr, "Running: %s\n", tmp);
  FILE *f = popen(tmp, "r");
  readpnm(f, channels);
  fclose(f);
}

Matrix<float> read_diff(int s, int l) {
  Matrix<float> images[8][3];
  for (int image = 0; image < 8; image++) {
    read_image(s, l, image, images[image]);
  }
  int xsize = images[0][0].xsize;
  int ysize = images[0][0].ysize;
  Matrix<float> ret(xsize, ysize);
  Matrix<float> eq(4, 8 * 3);
  Matrix<float> ans(1, 8 * 3);
  for (int y = 0; y < ysize; y++) {
    for (int x = 0; x < xsize; x++) {
      for (int image = 0; image < 8; image++) {
	for (int c = 0; c < 3; c++) {
	  int row = image * 3 + c;
	  ans.value(0, row) = images[image][c].value(x, y);
	  eq.value(0, row) = image & (1<<c) ? 1.0 : 0.0;
	  eq.value(c + 1, row) = 1.0;
	}
      }
      ret.value(x, y) = eq.solve_nnls(ans).value(0, 0);
    }
  }
  return ret;
}

void save_matrix(Matrix<float> data, FILE *f) {
  fprintf(f, "%d %d\n", data.xsize, data.ysize);
  for (int y = 0; y < data.ysize; y++) {
    for (int x = 0; x < data.xsize; x++) {
      fprintf(f, "%.18f ", data.value(x, y));
    }
    fprintf(f, "\n");
  }
}

int main(int argc, char **argv) {
  int strings = atoi(argv[1]);
  int lights = atoi(argv[2]);
  Matrix<float> tmp;
  for (int s = 0; s < strings; s++) {
    for (int l = 0; l < lights; l++) {
      Matrix<float> image = read_diff(s, l);
      fprintf(stderr, "Diff: %f\n", image.value(image.xsize/2, image.ysize/2));
      if (tmp.xsize == 0) {
	tmp = Matrix<float>(lights * strings, image.xsize * image.ysize);
      }
      for (int y = 0; y < image.ysize; y++) {
	for (int x = 0; x < image.xsize; x++) {
	  tmp.value(s * lights + l, y * image.xsize + x) = image.value(x, y);
	}
      }
    }
  }
  fprintf(stderr,"Solving...\n");
  save_matrix(tmp.invert_nnls(), stdout);
}
