#include <string>
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <memory.h>
#include "matrix.h"
#include "lights.h"

class YuvStream {
public:
  YuvStream(const string &filename) {
    cdiv = 2;
#if 0
    setenv("FILE", filename.c_str(), 1);
    system("mkfifo -m 660 /tmp/stream.yuv");
    system("mkfifo -m 660 /tmp/follow.yuv");
    system("mplayer -ao null -vo yuv4mpeg:file=/tmp/stream.yuv -zoom -vf scale -xy 120 -sws 9 \"$FILE\" &");
    system("mplayer /tmp/follow.yuv &");
    f = fopen("/tmp/stream.yuv", "r");
    out = fopen("/tmp/follow.yuv", "w");
    eof_reached = 0;
#else
    setenv("FILE", filename.c_str(), 1);
    f = popen("avconv -i \"$FILE\" -s 120x90 -an -f yuv4mpegpipe -", "r");
    out = NULL;
    eof_reached = 0;
#endif
    read_header();
  }

  int c() {
    int c = getc(f);
    if (out) putc(c, out);
    return c;
  }

  int getnum() {
    int x = 0;
    while (1) {
      int y = c();
      y -= '0';
      if (y < 0 || y > 9) return x;
      x *= 10;
      x += y;
    }
  }

  void read_header() {
    int k;
    if((k=c()) != 'Y' ||
       (k=c()) != 'U' ||
       (k=c()) != 'V' ||
       (k=c()) != '4' ||
       (k=c()) != 'M' ||
       (k=c()) != 'P' ||
       (k=c()) != 'E' ||
       (k=c()) != 'G' ||
       (k=c()) != '2')
      goto headerr;

    while(1)
    {
      switch(c())
      {
	case ' ':
	  continue;
	case 'W':
	  xsize = getnum();
	  break;
	case 'H':
	  ysize = getnum();
	  break;
	case 'C':
	  cformat = getnum();
	  switch (cformat) {
	    case 444:
	      cdiv = 1;
	      break;
	  }
	  break;

	case '\n': return;
	case EOF: goto headeof;
	default:
	  while(1) {
	    switch(c()) {
	      case '\n':
		return;
	      case ' ':
		break;
	      case EOF:
		goto headeof;
	      default:
		continue;
	    }
	    break;
	  }
      }
    }

  headerr:
    perror("**while reading header");
    exit(1);

  headeof:
    fprintf(stderr,"**End of file in header\n");
    exit(1);
  }

  void read_frame(Matrix<float> *r, Matrix<float> *g, Matrix<float> *b)
  {
    Matrix<float> y_comp(xsize, ysize);
    Matrix<float> u_comp(xsize / cdiv, ysize / cdiv);
    Matrix<float> v_comp(xsize / cdiv, ysize / cdiv);
    
    int k;
    struct frame *ret;
    if(eof_reached) {
      fprintf(stderr, "AT EOF!\n");
      return;
    }
    switch ((k=c())) {
      case 'F': break;
      case EOF: goto stream_end;
      default: goto frameerr;
    }
    if((k=c()) != 'R' ||
       (k=c()) != 'A' ||
       (k=c()) != 'M' ||
       (k=c()) != 'E')
      goto frameerr;

    while (1) {
      switch(k=c()) {
	case '\n': break;
	case EOF: goto frameeof;
	default: continue;
      }
      break;
    }
    for (int y = 0; y < ysize; y++) {
      for (int x = 0; x < xsize; x++) {
	y_comp.value(x, y) = c();
      }
    }
    for (int y = 0; y < ysize / cdiv; y++) {
      for (int x = 0; x < xsize / cdiv; x++) {
	u_comp.value(x, y) = c();
      }
    }
    for (int y = 0; y < ysize / cdiv; y++) {
      for (int x = 0; x < xsize / cdiv; x++) {
	v_comp.value(x, y) = c();
      }
    }
#if 0
    fprintf(stderr, "YUV<%f %f %f>\n",
	    y_comp.value(0,0),
	    u_comp.value(0,0),
	    v_comp.value(0,0));
#endif
    
    for (int y = 0; y < ysize; y++) {
      for (int x = 0; x < ysize; x++) {
	float Y = y_comp.value(x, y);
	float U = u_comp.value(x/cdiv, y/cdiv);
	float V= v_comp.value(x/cdiv, y/cdiv);
	r->value(x, y) = (1.164*(Y - 16) + 1.596*(V - 128))/256;
	g->value(x, y) = (1.164*(Y - 16) - 0.813*(V - 128) - 0.391*(U - 128))/256;
	b->value(x, y) = (1.164*(Y - 16) + 2.018*(U - 128))/256;
      }
    }
#if 0
    fprintf(stderr, "RGB<%f %f %f>\n",
	    r->value(0,0),
	    g->value(0,0),
	    b->value(0,0));
#endif
    return;

  frameeof:
    fprintf(stderr,"**EOF while reading frame\n");
    exit(1);

  stream_end:
    fprintf(stderr,"**EOF!\n");
    exit(0);

  frameerr:
    fprintf(stderr,"**Unexpected data while reading frame: %c (%d)\n",k,k);
    for (int i = 0; i < 100; i++) {
      fprintf(stderr, "k = %c ", c());
    }
    exit(1);
  }


private:
  FILE *f;
  FILE *out;
  int xsize, ysize;
  int cformat;
  int cdiv;
  int eof_reached;
};

Matrix<float> read_matrix(const string &name) {
  FILE *f = fopen(name.c_str(), "r");
  int xsize, ysize;
  fscanf(f, "%d %d", &xsize, &ysize);
  Matrix<float> data(xsize, ysize);
  for (int y = 0; y < data.ysize; y++) {
    for (int x = 0; x < data.xsize; x++) {
      fscanf(f, "%f", &data.value(x, y));
    }
  }
  return data;
}

int scalecolor(float n) {
  int x = float(n * 256);
  if (x > 255) x = 255;
  if (x < 0) x = 0;
  return x;
}


void readpnm(FILE *f, Matrix<float> *channels) {
  int format, xsize, ysize;
  int maxval = 1;
  if(fscanf(f, "P%d %d %d", &format, &xsize, &ysize)<3 ||
     format < 1 || format > 6 ||
     ( (format%3)!=1 && !fscanf(f, "%d",&maxval)))
  {
    fprintf(stderr,"solve: Not a PPM file\n");
    exit(1);
  }
  
  getc(f);

  channels[0] = Matrix<float>(xsize, ysize);
  channels[1] = Matrix<float>(xsize, ysize);
  channels[2] = Matrix<float>(xsize, ysize);

  int bit = 0; 
  int data = 0;
  for(int y=0; y<ysize; y++) {
    for(int x=0; x<xsize; x++) {
      int r, g, b;
      switch(format) {
	case 1:
	case 2:
	  fscanf(f,"%d",&g);
	  r = b = g;
	  break;

	case 3:
	  fscanf(f,"%d %d %d",&r, &g, &b);
	  break;

	case 4:
	  if(!bit) { data=getc(f); bit=8; }
	  r = b = g = !!(data & 128);
	  data<<=1;
	  bit--;
	  break;

	case 5:
	  r = b = g = getc(f);
	  break;

	case 6:
	  r = getc(f);
	  b = getc(f);
	  g = getc(f);
      }
      channels[0].value(x, y) = r / (float)maxval;
      channels[1].value(x, y) = r / (float)maxval;
      channels[2].value(x, y) = r / (float)maxval;
    }
  }
}

void read_image(int s, int l, int x, Matrix<float> *channels) {
  char tmp[256];
  sprintf(tmp, "djpeg -scale 1/8 -pnm images/light_%02d_%02d_%d.jpg",
	  s, l, x);
  fprintf(stderr, "Running: %s\n", tmp);
  FILE *f = popen(tmp, "r");
  readpnm(f, channels);
  fclose(f);
}

Matrix<float> read_diff(int s, int l) {
  Matrix<float> images[8][3];
  for (int image = 0; image < 8; image++) {
    read_image(s, l, image, images[image]);
  }
  int xsize = images[0][0].xsize;
  int ysize = images[0][0].ysize;
  Matrix<float> ret(xsize, ysize);
  Matrix<float> eq(4, 8 * 3);
  Matrix<float> ans(1, 8 * 3);
  for (int y = 0; y < ysize; y++) {
    for (int x = 0; x < xsize; x++) {
      for (int image = 0; image < 8; image++) {
	for (int c = 0; c < 3; c++) {
	  int row = image * 3 + c;
	  ans.value(0, row) = images[image][c].value(x, y);
	  eq.value(0, row) = image & (1<<c) ? 1.0 : 0.0;
	  eq.value(c + 1, row) = 1.0;
	}
      }
      ret.value(x, y) = eq.solve_nnls(ans).value(0, 0);
    }
  }
  return ret;
}

Matrix<float> read_solution(int lights, int strings) {
  Matrix<float> solution = read_matrix("solution");
  // Matrix<float> solution = tmp.invert_nnls();
  Matrix<float> tmp(120, 90);
  for (int x = 0; x < 120; x++)
    for (int y = 0; y < 90; y++)
      tmp.value(x,y) = 1.0;
  tmp.Reshape(1, tmp.xsize * tmp.ysize);
  tmp = solution * tmp;
  double sum = 0.0;
  for (int l = 0; l < lights; l++)
    for (int s = 0; s < strings; s++)
      sum += tmp.value(0, l + s * lights);
  return solution * (lights * strings / sum);
}

int main(int argc, char **argv) {
  int strings = atoi(argv[1]);
  int lights = atoi(argv[2]);

#if 0
  Matrix<float> tmp;
  for (int s = 0; s < strings; s++) {
    for (int l = 0; l < lights; l++) {
      Matrix<float> image = read_diff(s, l);
//      Matrix<float> image = read_image(s, l, 1);
      if (tmp.xsize == 0) {
	tmp = Matrix<float>(lights * strings, image.xsize * image.ysize);
      }
      for (int y = 0; y < image.ysize; y++) {
	for (int x = 0; x < image.xsize; x++) {
	  tmp.value(s * lights + l, y * image.xsize + x) = image.value(x, y);
	}
      }
    }
  }
#endif
  Matrix<float> r(120, 90);
  Matrix<float> g(120, 90);
  Matrix<float> b(120, 90);
  Matrix<float> solution = read_solution(lights, strings);

  YuvStream stream(argv[3]);

  TeensyLightWallRelay relay("", 8);
  relay.Init();
  while (1) {
    r.Reshape(120, 90);
    g.Reshape(120, 90);
    b.Reshape(120, 90);
    stream.read_frame(&r, &g, &b);
    r.Reshape(1, r.xsize * r.ysize);
    g.Reshape(1, g.xsize * g.ysize);
    b.Reshape(1, b.xsize * b.ysize);
#if 1
    Matrix<float> red_lights = solution * r;
    Matrix<float> green_lights = solution * g;
    Matrix<float> blue_lights = solution * b;
#else
    Matrix<float> red_lights = tmp.solve_nnls(r);
    Matrix<float> green_lights = tmp.solve_nnls(g);
    Matrix<float> blue_lights = tmp.solve_nnls(b);
#endif
#if 0
    fprintf(stderr, "RGB<%f %f %f>\n",
	    red_lights.value(0,0),
	    green_lights.value(0,0),
	    blue_lights.value(0,0));
#endif
    for (int l = 0; l < lights; l++) {
      vector<LampColor> colors(strings);
      for (int s = 0; s < strings; s++) {
	colors[s].intensity = 0xCC;
	colors[s].color[0] = scalecolor(red_lights.value(0, l + s * lights));
	colors[s].color[1] = scalecolor(green_lights.value(0, l + s * lights));
	colors[s].color[2] = scalecolor(blue_lights.value(0, l + s * lights));
      }
      relay.SetPixels(l, colors);
    }
  }
}
